import argparse

import torch
from ..esm.data import Alphabet
from argparse import ArgumentParser
from ..esm.model import ProteinBertModel, MSATransformer


def TinyEsm1b_config():
    alphabet = Alphabet.from_architecture('ESM-1b')
    parser = ArgumentParser()
    parser.add_argument(
        "--arch", default="roberta_large"
    )
    parser.add_argument(
        "--layers", default=12, type=int
    )
    parser.add_argument(
        "--max_positions", default=1024, type=int
    )
    parser.add_argument(
        "--embed_dim", default=640, type=int, metavar="N", help="embedding dimension"
    )
    parser.add_argument(
        "--emb_layer_norm_before", default=True, type=bool
    )
    parser.add_argument(
        "--logit_bias", action="store_true", help="whether to apply bias to logits"
    )
    parser.add_argument(
        "--ffn_embed_dim",
        default=2560,
        type=int,
        metavar="N",
        help="embedding dimension for FFN",
    )
    parser.add_argument(
        "--attention_heads",
        default=20,
        type=int,
        metavar="N",
        help="number of attention heads",
    )

    return parser.parse_args(), alphabet


def ESM1b_config():
    alphabet = Alphabet.from_architecture('ESM-1b')
    params = {'arch': 'roberta_large',
              'layers': 33,
              'max_positions': 1024,
              'embed_dim': 1280,
              'emb_layer_norm_before': True,
              'ffn_embed_dim': 5120,
              'attention_heads': 20}

    return argparse.Namespace(**params), alphabet


def MSA_config():
    alphabet = Alphabet.from_architecture('MSA Transformer')
    params = {'layers': 12,
              'embed_dim': 768,
              'ffn_embed_dim': 3072,
              'attention_heads': 12,
              'dropout': 0.1,
              'attention_dropout': 0.1,
              'activation_dropout': 0.1,
              'max_tokens': 2 ** 14,
              'max_positions': 1024,
              'embed_positions_msa': True}

    return argparse.Namespace(**params), alphabet


def load_model_architecture(name):
    if name == 'esm-1b':
        args, alphabet = ESM1b_config()
        model = ProteinBertModel(args, alphabet)
        return model

    elif name == 'msa':
        args, alphabet = MSA_config()
        model = MSATransformer(args, alphabet)
        return model

    elif name == 'esm-1b-tiny':
        args, alphabet = TinyEsm1b_config()
        model = ProteinBertModel(args, alphabet)
        return model


import pandas as pd
import os
import numpy as np
import platform
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from ..utils import TimeCounter

abs_path = os.path.dirname(os.path.dirname(__file__))
# use different commands for windows and linux respectively
system = platform.system().lower()
if system == 'linux':
    rm = 'rm {}'
    mv = 'mv {} {}'
    align_tool = f"{abs_path}/software/linux_famsa"
    command = align_tool + " {} {} >/dev/null 2>&1"

elif system == 'windows':
    rm = 'del {}'
    mv = 'move {} >nul 2>nul'
    align_tool = f"{abs_path}\\software\\windows_famsa.exe"
    command = align_tool + " {} {} >nul 2>nul"


def fasta2a3m(id, path):
    records = SeqIO.parse(path, "fasta")
    gap_id = ord('-')
    letter_gap = ord('A') - ord('a')
    
    for record in SeqIO.parse(path, "fasta"):
        if record.id == id:
            query_record = record
            seq = bytes(query_record.seq)
            seq_ids = np.array([id for id in seq])
            info = query_record.description
            # 查询序列中有'-'的位置
            ori_gaps = seq_ids == gap_id
            seq = ''.join(chr(id) for id in seq_ids[~ori_gaps])
    
    with open(f'{path}.a3m', 'w') as f:
        f.write(f'>{info}\n{seq}\n')
        for record in records:
            if record.id == query_record.id:
                continue
            
            seq = bytes(record.seq)
            info = record.description
            seq_ids = np.array([id for id in seq])
            
            # 同源序列中有'-'的位置
            seq_gaps = seq_ids == gap_id
            
            seq_ids[ori_gaps] = seq_ids[ori_gaps] - letter_gap
            
            mask = seq_gaps & ori_gaps
            seq = ''.join(chr(id) for id in seq_ids[~mask])
            
            f.write(f'>{info}\n{seq}\n')
    
    os.system(mv.format(f"{path}.a3m", path))


def build_msa(fasta_path, out_path, id, fmt='a3m'):
    os.system(command.format(fasta_path, out_path))
    if fmt == 'a3m':
        fasta2a3m(id, out_path)


def tsv_to_msa(tsv_path, output, align=True, fmt='a3m'):
    data = pd.read_csv(tsv_path, sep='\t')
    
    with TimeCounter('Generating files...'):
        for i, item in enumerate(data.groupby("query_id").indices.items()):
            query_id, indices = item
            info = data.iloc[indices].values
            
            out_name = f"{output}/query_{i}.faa" if output else f"query_{i}.faa"
            with open(out_name, 'w') as f:
                query_description, query_seq = info[0][1], info[0][4]
                f.write(f">{query_description}\n{query_seq}\n")
                
                for target in info:
                    f.write(f">{target[3]}\n{target[5]}\n")
            
            if align:
                build_msa(f"{out_name}", f"{out_name}_aligned", query_id, fmt=fmt)
                
                if fmt == 'a3m':
                    os.system(mv.format(f"{out_name}_aligned", f"{out_name[:-4]}.a3m"))
                    os.system(rm.format(out_name))
                elif fmt == 'fasta':
                    os.system(mv.format(f"{out_name}_aligned", out_name[:-4]))


def parse_tsv(tsv_path):
    data = pd.read_csv(tsv_path, sep='\t')
    
    for i, item in enumerate(data.groupby("query_id").indices.items()):
        query_id, indices = item
        info = data.iloc[indices].values
        
        query_description, query_seq = info[0][1], info[0][4]
        
        print(f"no.{i + 1}\nquery id: {query_id}")
        print(f"description: {query_description}")
        print(f"sequence length: {len(query_seq)}\n")
        
        print("detected sequences:")
        print(f"number:{len(info)}")
        print(f"{'target id':<40}{'length':<10}{'distance':<10}{'description':^50}")
        for target in info:
            hit_len = len(target[-2])
            hit_id = target[2]
            hit_description = target[3]
            distance = target[-1]
            print(f"{hit_id:<40}{hit_len:<10}{distance:.4f}\t{hit_description:<50}")
        
        print("*" * 100 + '\n\n')


def write_to_tsv(q_info_path, db_info_path, out_path, homo_list, dist_list):
    # 将索引排序
    index_set = set()
    for homo in homo_list:
        index_set = index_set | set(homo)
    sorted_list = sorted(list(index_set))
    
    # 读取向量索引库，获取索引对应的序列信息
    homo_dict = {}
    with open(db_info_path, 'r') as f:
        cnt = 0
        f.readline()
        if not sorted_list:
            print("No homology detected!")
            return
        
        for i in range(sorted_list[-1] + 1):
            data = f.readline()
            if i == sorted_list[cnt]:
                data = data.split('\t')
                homo_dict[i] = (data[0], data[1], data[2][:-1])
                cnt += 1
    
    # 输出查询结果
    with open(out_path, 'w') as f:
        f.write(f"query_id\tquery_description\ttarget_id\ttarget_description\tquery_seq\ttarget_seq\tdistance\n")
        q_info = pd.read_csv(q_info_path, sep='\t')
        for i, record in enumerate(q_info.values):
            query_id, query_description, query_seq = record
            
            if len(homo_list[i]) == 0:
                print(f"No homology detected for {query_id}!")
                continue
            
            for j, index in enumerate(homo_list[i]):
                target_id, target_description, target_seq = homo_dict[index]
                if j != 0:
                    query_seq = ""
                f.write(f"{query_id}\t{query_description}\t{target_id}\t"
                        f"{target_description}\t{query_seq}\t{target_seq}\t{dist_list[i][j]}\n")


if __name__ == '__main__':
    path = 'UniRef30_2020_06_900_1000_index.tsv'
    parse_tsv(path)


import faiss
import os
import numpy as np
import torch
import pandas as pd
from .model import MSARetrieveModel
from .utils import TimeCounter
from .vector_construction import fasta2vec
from .search import pipeline


def twintower_eval(q_res_path, db_test_path):
	res = pd.read_csv(q_res_path, sep='\t')
	db = pd.read_csv(db_test_path, sep='\t')

	info_dict = {}
	for _, row in db.iterrows():
		cls, seq = row
		info_dict[seq] = cls
		if cls in info_dict.keys():
			info_dict[cls] += 1
		else:
			info_dict[cls] = 1

	total_P = 0
	total_N = 0
	total_TP = 0

	for q_id, group in res.groupby('query_id'):
		TP = 0
		cls = info_dict[q_id]
		P = info_dict[cls]
		# N = min(group.shape[0], P)
		N = group.shape[0]
		# if P <= 100:
		# 	continue
		
		total_P += P
		total_N += N
		
		for i, data in enumerate(group.values):
			if i == N:
				break
			hit = data[2]
			if info_dict[q_id] == info_dict[hit]:
				total_TP += 1
				TP += 1

		recall = TP / P * 100
		precision = TP / N * 100 if N != 0 else 0

		print(f"query_ID:{q_id}\tSF_ID:{cls:<10}recall:{f'{TP}/{P}({recall:.2f}%)':<20}"
		      f"precision:{f'{TP}/{N}({precision:.2f}%)':<20}")

	print(f'recall: {total_TP}/{total_P}({total_TP/total_P*100:.2f}%)')
	print(f"precision: {total_TP}/{total_N}({total_TP/total_N*100:.2f}%)")


# makeblastdb -in test_db_scope2.08.faa -dbtype prot -input_type fasta -out /home/public/sj/data/scope2.08
# psiblast -query test_scope2.07.faa -db /home/public/sj/data/scope2.07 -out query_blast.tsv -outfmt "6 qseqid sseqid" -num_iterations 2
def blast_eval(res_path, db_path):
	data = pd.read_csv(res_path, sep='\t', header=None, names=['SCOP_ID', 'HOMO_ID'])
	db = pd.read_csv(db_path, sep='\t')
	data.dropna(inplace=True)

	total_TP = 0
	total_P = 0
	total_N = 0
	l = len(data.groupby('SCOP_ID').indices.keys())
	for query_id, homo_ids in data.groupby('SCOP_ID').indices.items():
		sf_id = db[db['SCOP_ID'] == query_id]['SF_ID'].values[0]
		pred_homo_ids = set(data['HOMO_ID'].iloc[homo_ids].values)
		true_homo_ids = set(db[db['SF_ID'] == sf_id]['SCOP_ID'].values)

		P = len(true_homo_ids)
		N = len(pred_homo_ids)
		TP = len(true_homo_ids & pred_homo_ids)

		recall = TP / P * 100
		precision = TP / N * 100

		total_TP += TP
		total_P += P
		total_N += N

		print(f"query_ID:{query_id}\tFOLD_ID:{sf_id:<10} recall:{f'{TP}/{P}({recall:.2f}%)':<20}"
		      f"precision:{f'{TP}/{N}({precision:.2f}%)':<20}")

	print(f'recall: {total_TP}/{total_P}({total_TP / total_P * 100:.2f}%)')
	print(f"precision: {total_TP}/{total_N}({total_TP / total_N * 100:.2f}%)")


# jackhmmer -N 1 --tblout query_hmmer test_scope2.07.faa test_db_scope2.07.faa
# hmmpy query_hmmer -o query_hmmer.tsv
def hmmer_eval(hmmer_path, db_path):
	data = pd.read_csv(hmmer_path, sep='\t')
	db = pd.read_csv(db_path, sep='\t')

	total_TP = 0
	total_P = 0
	total_N = 0
	indices = data.groupby('tlen').indices

	for query_id, homo_ids in indices.items():
		sf_id = db[db['SCOP_ID'] == query_id]['SF_ID'].values[0]
		pred_homo_ids = set(data['target name'].iloc[homo_ids].values)
		true_homo_ids = set(db[db['SF_ID'] == sf_id]['SCOP_ID'].values)

		P = len(true_homo_ids)
		N = len(pred_homo_ids)
		TP = len(true_homo_ids & pred_homo_ids)
		# if P <= 100:
		# 	continue
		recall = TP / P * 100
		precision = TP / N * 100

		total_TP += TP
		total_P += P
		total_N += N

		print(f"query_ID:{query_id}\tFOLD_ID:{sf_id:<10} recall:{f'{TP}/{P}({recall:.2f}%)':<20}"
		      f"precision:{f'{TP}/{N}({precision:.2f}%)':<20}")

	print(f'recall: {total_TP}/{total_P}({total_TP / total_P * 100:.2f}%)')
	print(f"precision: {total_TP}/{total_N}({total_TP / total_N * 100:.2f}%)")

# /mnt/beegfs/inspurfs/user-fs/sujin/dataset/mmseqs2/tools
def mmseqs2_eval(path, db_path):
	data = pd.read_csv(path, sep='\t', header=None, names=['qseqid', 'sseqid', 'pident', 'length', 'mismatch', 'gapopen',
	                                                       'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore'])

	db = pd.read_csv(db_path, sep='\t')
	total_TP = 0
	total_P = 0
	total_N = 0
	indices = data.groupby('qseqid').indices
	
	for query_id, homo_ids in indices.items():
		sf_id = db[db['SCOP_ID'] == query_id]['SF_ID'].values[0]
		pred_homo_ids = set(data['sseqid'].iloc[homo_ids].values)
		true_homo_ids = set(db[db['SF_ID'] == sf_id]['SCOP_ID'].values)
		
		P = len(true_homo_ids)
		N = len(pred_homo_ids)
		TP = len(true_homo_ids & pred_homo_ids)
		
		recall = TP / P * 100
		precision = TP / N * 100
		
		total_TP += TP
		total_P += P
		total_N += N
		
		print(f"query_ID:{query_id}\tFOLD_ID:{sf_id:<10} recall:{f'{TP}/{P}({recall:.2f}%)':<20}"
		      f"precision:{f'{TP}/{N}({precision:.2f}%)':<20}")
	
	print(f'recall: {total_TP}/{total_P}({total_TP / total_P * 100:.2f}%)')
	print(f"precision: {total_TP}/{total_N}({total_TP / total_N * 100:.2f}%)")
	

def hhblits_eval(hhblits_path, db_path):
	data = pd.read_csv(hhblits_path, sep='\t')
	db = pd.read_csv(db_path, sep='\t')

	total_TP = 0
	total_P = 0
	total_N = 0
	indices = data.groupby('query').indices
	l = len(indices.keys())

	for query_id, homo_ids in indices.items():
		sf_id = db[db['SCOP_ID'] == query_id]['SF_ID'].values[0]
		pred_homo_ids = set(data['target'].iloc[homo_ids].values)
		true_homo_ids = set(db[db['SF_ID'] == sf_id]['SCOP_ID'].values)

		P = len(true_homo_ids)
		N = len(pred_homo_ids)
		TP = len(true_homo_ids & pred_homo_ids)

		recall = TP / P * 100
		precision = TP / N * 100

		total_TP += TP
		total_P += P
		total_N += N

		print(f"query_ID:{query_id}\tFOLD_ID:{sf_id:<10} recall:{f'{TP}/{P}({recall:.2f}%)':<20}"
		      f"precision:{f'{TP}/{N}({precision:.2f}%)':<20}")

	print(f'recall: {total_TP}/{total_P}({total_TP / total_P * 100:.2f}%)')
	print(f"precision: {total_TP}/{total_N}({total_TP / total_N * 100:.2f}%)")


if __name__ == '__main__':
	device_ids = [0, 1, 2, 3]
	# model_name = "esm1b_t33_650M_UR50S"
	# model_name = "RetrieveModel_scope2.08_d1_m1_I13157_B256"
	# model_name = 'RetrieveModel_scope2.08_d1_m1_t12_I663_B256’
	model_name = 'RetrieveModel_scope2.08_d1_m1_t62_I663_B256_phase3'
	model_path = f"PretrainedModels/{model_name}.pt"
	q_seq_path = 'Data/test_scope2.08.faa'
	save_path = f'Faiss/test_{model_name}'
	fasta2vec(device_ids, model_path, q_seq_path, save_path, cover=False)
	
	q_vec_path = f'{save_path}.npy'
	q_info_path = f'{save_path}_index.tsv'
	name = '800_900'
	db_index_path = f"/sujin/dataset/uniclust30/UniRef30_2020_06_{name}.index"
	db_info_path = f"/sujin/dataset/uniclust30/UniRef30_2020_06_{name}_index.tsv"
	
	# 查询向量
	save_path = f'Results/UniRef30_2020_06_{name}_index.tsv'
	# pipeline(q_vec_path, db_index_path, q_info_path, db_info_path, save_path, iter_num=1, t=1, nprobe=10, max_num=None)
	
	# 评估双塔模型
	res_path = "Faiss/test_RetrieveModel_scope2.08_d1_m1_t60_I100_B256_phase3_result.tsv"
	q_test_path = "test_scope2.08.tsv"
	db_test_path = "Data/db_scope2.08.tsv"
	twintower_eval(res_path, db_test_path)
	
	db_path = 'Data/test_db_scope2.08.tsv'
	# 评估hmmer3
	hmmer_path = 'Data/query_hmmer.tsv'
	# hmmer_eval(hmmer_path, db_path)
	
	# evaluate mmseqs2
	mmseqs2_path = "Data/resultDB.m8"
	# mmseqs2_eval(mmseqs2_path, db_path)
	
	# evaluate hhblits
	hhblits_path = "Data/query_hhblits.tsv"
	# hhblits_eval(hhblits_path, db_path)
	
	# 评估psi-blast
	blast_path = 'Data/query_blast.tsv'
	# blast_eval(blast_path, db_path)


import re
import random
import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from tqdm import tqdm
from ..utils import TimeCounter


# 过滤长度过长的序列
def data_filter(path, out_path, threshold=512):
	with open(out_path, 'w') as f:
		for seq_record in tqdm(SeqIO.parse(path, "fasta")):
			if len(seq_record.seq) > threshold:
				continue
			else:
				f.write(f">{seq_record.description}\n")
				f.write(f"{str(seq_record.seq)}\n")


# SCOP2
def load_data(path):
	sf = {}
	pattern = r'SF=([\d]+)'

	# 提取文件里的所有序列，按照超家族分类
	for seq_record in SeqIO.parse(path, "fasta"):
		sf_id = re.search(pattern, seq_record.description).group(1)

		if sf_id not in sf.keys():
			sf[sf_id] = [seq_record]
		else:
			sf[sf_id].append(seq_record)

	# 去掉所有同源序列数量为1的超家族
	del_ids = []
	for sf_id, seqs in sf.items():
		if len(seqs) == 1:
			del_ids.append(sf_id)
	[sf.pop(key) for key in del_ids]

	return sf


#SCOPe v2.08
def load_scope_data(path):
	pattern = r' (.*?\d+.\d+)'
	records = SeqIO.parse(path, 'fasta')
	sf_dict = {}

	for seq in records:
		fd_id = re.search(pattern, str(seq.description)).group(1)
		seq.seq = Seq(str(seq.seq).upper())
		if fd_id not in sf_dict.keys():
			sf_dict[fd_id] = [seq]
		else:
			sf_dict[fd_id].append(seq)

	sf_dict = {k: v for k, v in sf_dict.items() if len(v) > 1}

	return sf_dict


# uniclust30_2018_08_seed
def sample_seqs(fasta_path, pointer, n, threshold):
	# total number of sequences
	total = pointer.shape[0]
	
	# sample n negative samples
	indices = [random.randint(0, total-1) for _ in range(n)]
	
	samples = []
	with open(fasta_path, 'r') as f:
		for index in indices:
			loc = pointer[index][0]
			
			# ensure length of sequence less than threshold
			while True:
				f.seek(loc)
				f.readline()
				seq = f.readline()[:-1]
				if len(seq) <= threshold:
					samples.append(seq)
					break
					
				else:
					new_index = random.randint(0, total-1)
					loc = pointer[new_index][0]
	
	return samples
			

# 对fasta文件构建指针索引
def construct_pointer(fasta_path, save_path):
	with TimeCounter("Constructing index file..."):
		with open(save_path, 'w') as w:
			w.write("location\n")
			with open(fasta_path, 'r') as r:
				loc = r.tell()
				line = r.readline()
				while line:
					if line[0] == '>':
						w.write(f'{loc}\n')

					loc = r.tell()
					line = r.readline()


# 将数据划分为训练集和测试集
def data_split(data):
	sorted_sf_dict = sorted(data.items(), key=lambda kv: (len(kv[1]), kv[0]), reverse=True)

	test_seqs = []
	test_db = []
	# 系统抽样
	for i in range(10, len(data), 10):
		key = sorted_sf_dict[i][0]
		seqs = data.pop(key)
		test_seqs.append(seqs[0])
		test_db += seqs

	return data, test_seqs, test_db


def training_construction(data):
	n_per_sf = 2000
	
	fasta_path = "/sujin/dataset/uniclust30/uniclust30_2018_08_seed.fasta"
	pointer_path = "/sujin/dataset/uniclust30/uniclust30_2018_08_seed.tsv"
	pointer = pd.read_csv(pointer_path, sep='\t').values

	with open("../Data/train_scope2.08.tsv", 'w') as f:
		f.write("ori\tpos\tneg\n")

		for records in tqdm(data.values()):
			neg_seqs = sample_seqs(fasta_path, pointer, n_per_sf, 512)
			for i in range(n_per_sf):
				ori, pos = random.sample(records, 2)
				ori_seq = str(ori.seq)
				pos_seq = str(pos.seq)
				neg_seq = neg_seqs[i]
		
				f.write(f'{ori_seq}\t{pos_seq}\t{neg_seq}\n')


def threshold_data_construction(sf):
	with open("../Data/thresold_data_scope2.07.tsv", 'w') as f:
		f.write("seq\thomo_num\n")
		for k, v in sf.items():
			for seq in v:
				f.write(f"{str(seq.seq)}\t{len(v)-1}\n")


def create_tsv(data, path):
	pattern = r' (.*?\d+.\d+)'
	seqs = []
	# 将向量的顺序与id对应起来输出为索引文件
	with open(path, 'w') as f:
		f.write("SF_ID\tSCOP_ID\n")
		for i, record in enumerate(data):
			scop_id = record.id
			sf_id = re.search(pattern, record.description).group(1)
			seqs.append(str(record.seq))

			f.write(f"{sf_id}\t{scop_id}\n")


if __name__ == '__main__':
	# 文件的位置
	path = "../Data/scope2.08_<=512.faa"
	# out_path = "../Data/scope2.08_<=512.faa"
	# data_filter(path, out_path, threshold=512)
	# 读取文件
	sf_dict = load_scope_data(path)
	# 划分训练集测试集
	train_sf, test_seqs, test_db = data_split(sf_dict)
	# 将测试集输出为fasta文件
	SeqIO.write(test_seqs, "../Data/test_scope2.08.faa", "fasta")
	SeqIO.write(test_db, "../Data/test_db_scope2.08.faa", "fasta")
	# 构建测试tsv文件
	create_tsv(test_seqs, "../Data/test_scope2.08.tsv")
	create_tsv(test_db, "../Data/test_db_scope2.08.tsv")
	# 构建训练集
	training_construction(train_sf)
	
	# fasta_path = "/sujin/dataset/uniclust30/uniclust30_2018_08_seed.fasta"
	# pointer_path = "/sujin/dataset/uniclust30/uniclust30_2018_08_seed_pointer.tsv"
	# construct_pointer(fasta_path, pointer_path)
	# negs = sample_seqs(fasta_path, pointer_path, 100)
	# print(negs)


from setuptools import setup, find_packages
setup(
    name="twintowers",
    version="0.2.1",
    author="LTEnjoy",
    author_email="sujinltenjoy@gmail.com",
    description="protein homology detection using deep learning",
    install_requires=['faiss'],
    include_package_data=True,
    
    # 你要安装的包，通过 setuptools.find_packages 找到当前目录下有哪些包
    packages=find_packages(),
    
    entry_points={
        'console_scripts': [
            'twintowers = TwinTowers.command:tasks'
        ]
    }
)



import sys

import faiss
import os
import numpy as np
import torch
from math import ceil
from .model import MSARetrieveModel
from .utils import TimeCounter
from .vector_construction import fasta2vec
from .Results.format import *


def search(q_vec_path, db_index_path, iter_num=2, t=1, nprobe=1, max_num=500, verbose=True):
	
	with TimeCounter(f"Loading {db_index_path}...", verbose):
		index = faiss.read_index(db_index_path)
		
		if max_num == 0:
			max_num = index.ntotal
		
		index.nprobe = index.ntotal if nprobe == 0 else nprobe
			
	q_vec = np.load(q_vec_path)
	with TimeCounter("Searching index...", verbose):
		dist, indices = index.search(q_vec, max_num)
		dist = np.sqrt(dist)
	
	if iter_num > 1:
		db_vectors = index.reconstruct_n(0, index.ntotal)

		for times in range(iter_num-1):
			cnt = np.sum(dist <= t, axis=1)
			for i, n in enumerate(cnt):
				if n != 0:
					num = min(n, 100)
					homo_vecs = db_vectors[indices[i][:num]]
					q_vec[i] = (np.sum(homo_vecs, axis=0) + q_vec[i]) / (homo_vecs.shape[0] + 1)

			with TimeCounter(f"Iteratively searching index...{times+2}", verbose):
				dist, indices = index.search(q_vec, max_num)
				dist = np.sqrt(dist)

	cnt = np.sum(np.bitwise_and(dist <= t, dist >= 0), axis=1)
	homo_list = [indices[i, :n] for i, n in enumerate(cnt)]
	dist_list = [dist[i, :n] for i, n in enumerate(cnt)]

	return homo_list, dist_list


def pipeline(q_vec_path, db_index_path, q_info_path, db_info_path, save_path, iter_num=1, t=1, nprobe=1, max_num=500, verbose=True):
	homo_list, dist_list = search(q_vec_path, db_index_path, iter_num, t, nprobe, max_num, verbose)
	with TimeCounter("output results...", verbose):
		write_to_tsv(q_info_path, db_info_path, save_path, homo_list, dist_list)


if __name__ == '__main__':
	# 对查询序列构建索引库和向量库
	device_ids = [0, 1, 2, 3]
	model_path = "PretrainedModels/RetrieveModel_scope2.07_p1_n5_m4_I17100_B256.pt"
	q_seq_path = 'Data/test_scope2.07.faa'
	save_path = 'Faiss/test_scope2.07'
	# fasta2vec(device_ids, model_path, q_seq_path, save_path)

	q_vec_path = f'{save_path}.npy'
	q_info_path = f'{save_path}_index.tsv'
	db_index_path = "Faiss/uniclust30_2018_08_seed_Flat.index"
	db_info_path = "Faiss/uniclust30_2018_08_seed_index.tsv"

	# 查询向量
	save_path = 'Results/test_scope2.07_res_Flat.tsv'
	# pipeline(q_vec_path, db_index_path, q_info_path, db_info_path, save_path, iter_num=2, t=1, gpu=False)

	msa_path = "Results"
	name = "query"
	tsv_to_msa(save_path, msa_path, name)



import torch
import os
import numpy as np
import logging
import multiprocessing as mp
from Bio import SeqIO
from math import ceil
from .LoadingData.tokenizer import Tokenizer
from .model import MSARetrieveModel
from .utils import progress_bar


# 利用多进程的方式获取数据库序列高维向量
def mp_get_vec(device_ids, model, seqs, save_path=None, batch_size=128, verbose=True):
    torch.multiprocessing.set_start_method('spawn')
    # 将数据分割为多份
    n = ceil(len(seqs) / len(device_ids))
    p_list = []
    for i, id in enumerate(device_ids):
        if save_path:
            path = save_path + str(i)
        else:
            path = 'temp' + str(i)
        seq_list = seqs[n * i: n * i + n]
        p = mp.Process(target=get_vec, args=(id, model, seq_list, path, batch_size, verbose))
        p.start()
        p_list.append(p)

    # 等待所有进程执行完毕
    for p in p_list:
        p.join()

    # 整合生成的向量数据
    vectors = np.empty((len(seqs), model.dim), dtype=np.float32)
    pointer = 0
    for i in range(len(device_ids)):
        if save_path:
            path = save_path + str(i) + ".npy"
        else:
            path = 'temp' + str(i) + ".npy"

        v = np.load(path)
        os.remove(path)
        vectors[pointer: pointer + v.shape[0]] = v
        pointer += v.shape[0]

    if save_path:
        np.save(save_path, vectors)

    return vectors


# 获取数据库序列的高维向量
def get_vec(device, model, seqs, save_path, batch_size=128, verbose=True):
    tokenizer = Tokenizer()
    model.to(device)
    model.eval()
    vectors = np.empty((len(seqs), model.dim), dtype=np.float32)
    with torch.no_grad():
        for i in range(0, len(seqs), batch_size):
            slice = seqs[i: i + batch_size]
            info = tokenizer.batch_encode(slice, padding=True)
            token_ids = info['input_ids']
            lengths = info['lengths']
            inputs = token_ids.to(device)
            outputs = model(inputs)
            repr = outputs['repr']
            # t = outputs['t'].unsqueeze(-1)
            vectors[i: i + batch_size] = repr.to('cpu')

            desc = f"device:{device} calculating vectors"
            if verbose:
                progress_bar(i + batch_size, len(seqs), desc)

    if save_path:
        np.save(save_path, vectors)

    return vectors


# 构建向量库索引
def construct_index(records, path):
    seqs = []
    with open(path, 'w') as f:
        f.write("id\tdecription\tsequence\n")
        for record in records:
            if len(record.seq) > 1022:
                print(f"sequence {record.id} has length > 1022, so ignored.")
                continue
            seqs.append(str(record.seq))
            f.write(f"{record.id}\t{record.description}\t{str(record.seq)}\n")

    return seqs


# 将fasta序列转化为numpy的向量
def fasta2vec(device_ids, model_path, fasta_path, save_path, batch_size=128, log_name=None, cover=False, verbose=True):
    if os.path.exists(f"{save_path}.npy") and not cover:
        if verbose:
            print("npy file already exists.")
        return
    
    records = SeqIO.parse(fasta_path, 'fasta')
    seqs = construct_index(records, f"{save_path}_info.tsv")

    model = MSARetrieveModel()
    model.load_state_dict(torch.load(model_path, map_location='cpu'))

    return mp_get_vec(device_ids, model, seqs, save_path=save_path, batch_size=batch_size, verbose=verbose)
